$WANDERER_PLACE_RADIUS = 24;
$WANDERER_PLACE_CHANCE = 28; // 24

function ZAMB_Director(%zamb) {
	return new ScriptObject() {
		class = ZAMB_Director;
		zamb = %zamb;
	};
}

function ZAMB_Director::onAdd(%this) {
	%this.pendingWandererSet = new SimSet();

	for (%i = NodeGroup.getCount() - 1; %i >= 0; %i--) {
		%node = NodeGroup.getObject(%i);

		if (mFloor(getRandom() * 2.5 / (%node.nearNodes / $WANDERER_PLACE_CHANCE))) {
			%this.pendingWandererSet.add(%node);
		}
	}
}

function ZAMB_Director::onRemove(%this) {
	%this.pendingWandererSet.delete();
}

function ZAMB_Director::tick(%this) {
	%this.spawnWanderers();
	%this.spawnHordes();
	%this.spawnSpecials();
	%this.spawnBosses();
}

// Main tick functions

function ZAMB_Director::spawnWanderers(%this) {
	%count = %this.zamb.zombies.wanderers.getCount();
	%range = %this.zamb.zombies.wanderers.spawnRange;
	%limit = %this.zamb.zombies.wanderers.limit;

	%i = %this.pendingWandererSet.getCount() - 1;

	for (%i; %i >= 0 && %count < %limit; %i--) {
		%node = %this.pendingWandererSet.getObject(%i);

		for (%j = 0; %j < %this.zamb.miniGame.numMembers; %j++) {
			%obj = %this.zamb.miniGame.member[%j].player;

			if (isObject(%obj) && %obj.getState() !$= "Dead") {
				if (vectorDist(%obj.position, %node.position) <= %range) {
					break;
				}
			}
		}

		if (%j == %this.zamb.miniGame.numMembers) {
			continue;
		}

		if (%this.zamb.zombies.spawn(WandererZombieData, %node) != -1) {
			%count++;
			%this.pendingWandererSet.remove(%node);
		}
	}
}

function ZAMB_Director::spawnHordes(%this) {
	if (%this.nextHorde $= "") {
		%this.nextHorde = $Sim::Time + %this.getHordeSpawnInterval();
	}

	if (%this.nextHorde <= $Sim::Time) {
		%this.horde();
	}
}

function ZAMB_Director::spawnSpecials(%this) {
	if ($Sim::Time - %this.zamb.start < 30) {
		return;
	}

	if (%this.zamb.zombies.specials.getCount() >= %this.zamb.zombies.specials.limit) {
		return;
	}

	%types = "boomerZombieData";
	%count = %this.zamb.zombies.specials.getCount();

	for (%i = 0; %i < %count; %i++) {
		%obj = %this.zamb.zombies.specials.getObject(%i);
		%exists[%obj.getDataBlock()] = 1;
	}

	%count = getFieldCount(%types);

	for (%i = 0; %i < %count; %i++) {
		%type = getField(%types, %i).getID();

		if (%exists[%type]) {
		 	continue;
		}

		%object = %this.specialObject[%type];

		if (!isObject(%object) || %object.getState() $= "Dead") {
			if (%this.specialTimer[%type] $= "") {
				%this.specialTimer[%type] = $Sim::Time + %this.getSpecialSpawnInterval();
			}

			if (%this.specialTimer[%type] <= $Sim::Time) {
				if (%valid $= "") {
					%valid = %type;
				}
				else {
					%valid = %valid SPC %type;
				}
			}
		}
	}

	%type = pickRandomItemFromList(%valid);

	if (%type !$= "") {
		%obj = %this.zamb.zombies.spawn(%type);

		if (isObject(%obj)) {
			%this.specialObject[%type] = %obj;
			%this.specialTimer[%type] = "";
		}
	}
}

function ZAMB_Director::spawnBosses(%this) {
	if (%this.nextBoss $= "") {
		%this.nextBoss = $Sim::Time + %this.getBossSpawnInterval();
	}

	if (%this.nextBoss <= $Sim::Time) {
		%this.boss();
	}
}

// Support functions for spawning

function ZAMB_Director::horde(%this, %noSound) {
	%size = getRandom(10, 30);
	%node = NodeGroup.findZombieSpawn(%this.zamb.zombies.hordes.spawnRange);

	if (!%noSound) {
		if (%size < 20) {
			// serverPlay2D(zamb_music_germs_low);
		}
		else {
			// serverPlay2D(zamb_music_germs_high);
		}
	}

	%this.prevHorde = $Sim::Time;
	%this.nextHorde = $Sim::Time + %this.getHordeSpawnInterval();

	%this.hordeSpawner(%node, %size);
}

function ZAMB_Director::hordeSpawner(%this, %node, %size) {
	if (!isObject(%node) || !%size) {
		return;
	}

	%obj = %this.zamb.zombies.spawn(hordeZombieData, %node);

	if (isObject(%obj)) {
		// %obj.setRage(1);

		if (%size >= 2) {
			%this.schedule(1000, hordeSpawner, %node, %size - 1);
		}
	}
}

function ZAMB_Director::boss(%this, %type) {
	%this.nextBoss = "";

	if (%type $= "") {
		%types = "none tank";

		if (%this.lastBossType !$= "") {
			%types = trim(strReplace(" " @ %types @ " ", " " @ %this.lastBossType @ " ", " "));
		}

		%type = pickRandomItemFromList(%types);
	}

	if (%type $= "" || %type $= "none") {
		return;
	}

	switch$ (%type) {
		case "tank": %this.zamb.zombies.spawn(tankZombieData);
	}
}

function ZAMB_Director::getHordeSpawnInterval(%this) {
	return (4 - %this.zamb.difficulty) * 60 + getRandom(-20, 30);
}

function ZAMB_Director::getSpecialSpawnInterval(%this) {
	return getRandom(30, 60);
}

function ZAMB_Director::getBossSpawnInterval(%this) {
	return getRandom(150, 300);
}
